# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import json
import logging
import os
import sys
from pathlib import PosixPath

from dopplerr.cfgtree import get_node_by_xpath
from dopplerr.cfgtree import set_node_by_xpath

log = logging.getLogger(__name__)
_undefined = object()


class _CfgBase(object):

    _DEFAULT = None
    name = None
    xpath = None
    arg_type = None
    environ_var_prefix = None
    ignore_in_cfg = False

    def __init__(self, l=None, s=None, h=None, r=False, d=_undefined):
        # Note: self.name should come later by EnvironmentConfig._inject_names()
        self.short_param = s
        self.help_str = h
        self.required = r
        self.forced_long_param = l
        if d == _undefined:
            self.default = self._DEFAULT
        else:
            self.default = d
        self.value = self.default

    def set_value(self, value):
        """
        Setter method used in set_node_by_xpath
        """
        self.value = value

    @property
    def environ_var_name(self):
        return self.environ_var_prefix + self.cmd_line_name.upper()

    def get_cmd_line_params(self):
        a = []
        if self.short_param:
            a.append(self.short_param)
        if self.name:
            a.append(self.long_param)
        return a

    @property
    def _environ_var_value(self):
        return os.environ.get(self.environ_var_name, _undefined)

    def read_environ_var(self):
        return str(self._environ_var_value)

    @property
    def long_param(self):
        if self.forced_long_param:
            return self.forced_long_param
        if not self.xpath:
            return "--" + self.name.lower().replace("_", "-")
        return "--" + self.xpath.replace('.', '-').replace('_', '-')

    @property
    def cmd_line_name(self):
        return self.xpath.lower().replace("-", "_").replace(".", "_")

    @property
    def action(self):
        return 'store'

    @property
    def n_args(self):
        return None

    @property
    def safe_value(self):
        """
        string that can be outputed to logs
        """
        return self.value

    @property
    def cfgfile_value(self):
        """
        value as it should be saved in config file
        """
        return self.value if self.value is not None else ""

    @property
    def metavar(self):
        return self.name.upper()


class StringCfg(_CfgBase):
    _DEFAULT = ""

    def read_environ_var(self):
        return str(self._environ_var_value)


class ListOfStringCfg(_CfgBase):
    """
    Comma separated list of string (1 argument)
    """

    def __init__(self, *args, **kwargs):
        self._DEFAULT = []
        super(ListOfStringCfg, self).__init__(*args, **kwargs)

    def read_environ_var(self):
        ls = self._environ_var_value
        return ls.split(",")

    @property
    def cfgfile_value(self):
        """
        value as it should be saved in config file
        """
        return ",".join(self.value)

    @staticmethod
    def arg_type(string):
        return string.split(",")


class IntCfg(_CfgBase):
    _DEFAULT = 0

    def read_environ_var(self):
        return int(self._environ_var_value)


class UserCfg(StringCfg):
    @property
    def user(self):
        return self.value


class PasswordCfg(StringCfg):
    @property
    def password(self):
        return self.value

    @property
    def safe_value(self):
        """
        Hide password in logs
        """
        return "*" * len(self.value)


class DirNameCfg(StringCfg):
    _DEFAULT = None


class ConfigFileCfg(StringCfg):
    _DEFAULT = None
    ignore_in_cfg = True


class BoolCfg(_CfgBase):
    _DEFAULT = False

    def read_environ_var(self):
        e = os.environ.get(self.environ_var_name)
        return bool(e)

    @property
    def action(self):
        return 'store_true'

    @property
    def metavar(self):
        return None


class MultiChoiceCfg(ListOfStringCfg):
    def __init__(self, choices=None, *args, **kwargs):
        super(MultiChoiceCfg, self).__init__(*args, **kwargs)
        self.choices = choices

    def arg_type(self, string):
        items = string.split(",")
        for item in items:
            if item not in self.choices:
                raise argparse.ArgumentTypeError("{!r} not in available choise: {}".format(
                    item, ", ".join(self.choices)))


class EnvironmentConfig(object):
    _CFG = None
    _ENVIRON_VAR_PREFIX = None
    _CONFIGURATION_ENTRY_POINT = None

    def __init__(self):
        self._inject_names()

    @staticmethod
    def mkxpath(xpath, name):
        return xpath + "." + name if xpath else name

    def _inject_names(self, root=None, xpath=None):
        """
        Inject configuration item name defined in the _CFG dict inside each _Cfg
        """
        if root is None:
            if self._CFG is None:
                return
            root = self._CFG
        # pylint: disable=no-member
        for name, item in root.items():
            if isinstance(item, dict):
                self._inject_names(root=item, xpath=self.mkxpath(xpath, name))
            else:
                item.name = name
                item.xpath = self.mkxpath(xpath, name)
                item.environ_var_prefix = self._ENVIRON_VAR_PREFIX
                if item.ignore_in_cfg:
                    # log.debug("Create cfg node '%s': ignored (handled later)", item.xpath)
                    continue
                log.debug("Create cfg node: '%s' (name: '%s') with value: %r", item.xpath,
                          item.name, item.safe_value)
        # pylint: enable=no-member

    def set_cfg_value(self, xpath, value):
        """
        Set a value in _CFG
        """
        set_node_by_xpath(self._CFG, xpath, value, extend=True, setter_attr="set_value")

    def get_cfg_value(self, xpath, default=None):
        """
        Get a value from _CFG
        """
        return get_node_by_xpath(self._CFG, xpath, default=default).value

    def find_configuration_values(self):
        self._load_configuration()
        self._load_environment_variables("", self._CFG)
        self._load_cmd_line_arg()
        self._save_configuration()

    def _load_configuration(self):
        log.debug("Looking for configuration")
        self._CONFIGURATION_ENTRY_POINT.find_entry_point()
        bare_cfg = self._CONFIGURATION_ENTRY_POINT.get_bare_config_dict()
        self._load_cfg_dict(bare_cfg)

    def _save_configuration(self):
        log.debug("Saving configuration")
        bare_cfg = self._dict(safe=False)
        self._CONFIGURATION_ENTRY_POINT.save_bare_config_dict(bare_cfg)

    def _load_cfg_dict(self, cfg, xpath=None):
        for k, v in cfg.items():
            if isinstance(v, dict):
                self._load_cfg_dict(v, self.mkxpath(xpath, k))
            else:
                self.set_cfg_value(self.mkxpath(xpath, k), v)

    def _load_environment_variables(self, xpath, root):
        """
        Inject value from environment variable
        """
        for name, item in root.items():
            if isinstance(item, dict):
                self._load_environment_variables(self.mkxpath(xpath, name), item)
            else:
                print("item.environ_var_name:", item.environ_var_name)
                if item.environ_var_name in os.environ:
                    if item.ignore_in_cfg:
                        log.debug("Ignoring environment variable %s", item.environ_var_name)
                    val = item.read_environ_var()
                    log.debug("Found environment variable '%s': %s (conf: %s)",
                              item.environ_var_name, val, item.xpath)
                    item.value = val

    def _load_cmd_line_arg(self):
        """
        Inject parameters provider by the user in the command line
        """
        argv = sys.argv[1:]

        parser = argparse.ArgumentParser()
        self._inject_cfg_in_parser(parser)
        args = parser.parse_args(args=argv)
        for k, v in vars(args).items():
            cfg = self._find_cfg_for_cmd_line_name(k)
            if v is _undefined:
                continue
            cfg.value = v
            if cfg.ignore_in_cfg:
                log.debug("Ignoring command line parameter %s", cfg.long_param)
            log.debug("Found command line parameter '%s': %s (conf: %s)", cfg.long_param,
                      cfg.safe_value, cfg.xpath)
        log.debug("Current configuration: %s", self._json(safe=True))

    def _find_cfg_for_cmd_line_name(self, cmd_line_name, root=None):
        if root is None:
            root = self._CFG
        for v in root.values():
            if isinstance(v, dict):
                f = self._find_cfg_for_cmd_line_name(cmd_line_name, root=v)
                if f:
                    return f
            else:
                if v.cmd_line_name == cmd_line_name:
                    return v

    def _inject_cfg_in_parser(self, parser, xpath=None, root=None):
        """
        Configure the argument parser according to _CFG
        """
        if root is None:
            root = self._CFG
        # pylint: disable=no-member
        for name, item in root.items():
            if isinstance(item, dict):
                self._inject_cfg_in_parser(parser, xpath=self.mkxpath(xpath, name), root=item)
            else:
                args = item.get_cmd_line_params()
                kwargs = {
                    "action": item.action,
                    "dest": item.cmd_line_name,
                    "help": item.help_str,
                    "default": _undefined,
                }
                # log.debug("parser arg %s, dest %s", item.long_param, kwargs['dest'])
                nargs = item.n_args
                if nargs:
                    kwargs["nargs"] = nargs
                metavar = item.metavar
                if metavar:
                    kwargs["metavar"] = metavar
                if item.arg_type is not None:
                    kwargs["type"] = item.arg_type
                parser.add_argument(*args, **kwargs)
        # pylint: enable=no-member

    def _dict(self, root=None, safe=False):
        """
        Return the configuration as a dictionnary
        """
        if root is None:
            root = self._CFG
        d = {}
        # pylint: disable=no-member
        for name, item in root.items():
            if isinstance(item, dict):
                d[name] = self._dict(root=item, safe=safe)
            else:
                if item.ignore_in_cfg:
                    continue
                elif safe:
                    d[name] = item.safe_value
                else:
                    d[name] = item.value
        # pylint: enable=no-member
        return d

    def _json(self, safe=False):
        return json.dumps(self._dict(safe=safe), sort_keys=True, indent=4, separators=(',', ': '))


class _ConfigEntryPointBase(object):
    def find_entry_point(self):
        raise NotImplementedError

    def get_bare_config_dict(self):
        raise NotImplementedError

    def save_bare_config_dict(self, bare_cfg):
        raise NotImplementedError


class JsonFileConfigEntryPoint(_ConfigEntryPointBase):
    _CONFIG_ENTRYPOINY_DEFAULT_FILENAME = None
    _CONFIG_ENTRYPOINT_ENVIRON_VAR_NAME = None
    _CONFIG_ENTRYPOINT_SHORT_PARAM_NAME = None
    _CONFIG_ENTRYPOINT_LONG_PARAM_NAME = None

    _resolved_config_file = None
    _bare_config_dict = None

    def find_entry_point(self):
        configfile = self._CONFIG_ENTRYPOINY_DEFAULT_FILENAME
        if self._CONFIG_ENTRYPOINT_ENVIRON_VAR_NAME in os.environ:
            configfile = os.environ[self._CONFIG_ENTRYPOINT_ENVIRON_VAR_NAME]
            log.debug("%s defined: %s", self._CONFIG_ENTRYPOINT_ENVIRON_VAR_NAME, configfile)
        for i in range(len(sys.argv)):
            good = []
            if self._CONFIG_ENTRYPOINT_SHORT_PARAM_NAME:
                good.append(self._CONFIG_ENTRYPOINT_SHORT_PARAM_NAME)
            if self._CONFIG_ENTRYPOINT_LONG_PARAM_NAME:
                good.append(self._CONFIG_ENTRYPOINT_LONG_PARAM_NAME)
            if sys.argv[i] in good:
                if i == len(sys.argv):
                    raise Exception("No value given to {}".format(" or ".join(good)))
                configfile = sys.argv[i + 1]
                log.debug("%s defined: %s", " or ".join(good), configfile)
                break
        config_file_path = PosixPath(configfile)
        log.info("Configuration file set to: %s", configfile)
        self._resolved_config_file = config_file_path.resolve().as_posix()
        self._load_bare_config()

    def _load_bare_config(self):
        log.debug("(Re)loading configuration file: %s", self._resolved_config_file)
        config_file_path = PosixPath(self._resolved_config_file)
        if config_file_path.exists():
            with config_file_path.open() as f:
                self._bare_config_dict = json.load(f)
        else:
            self._bare_config_dict = {}

    def save_bare_config_dict(self, bare_cfg):
        with PosixPath(self._resolved_config_file).open('w') as f:
            f.write(json.dumps(bare_cfg, sort_keys=True, indent=4, separators=(',', ': ')))

    def get_bare_config_dict(self):
        return self._bare_config_dict



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

from dopplerr.cfg import DopplerrConfig

log = logging.getLogger(__name__)


class _FilterBase(object):
    def appy_path_mapping(self, folder):
        if not DopplerrConfig().get_cfg_value("general.mapping"):
            return folder
        if folder.startswith("/"):
            absolute = True
            folder = folder[1:]
        for mapping in DopplerrConfig().get_cfg_value("general.mapping"):
            log.debug("Mapping: %s", mapping)
            k, _, v = mapping.partition("=")
            log.debug("Applying mapping %s => %s", k, v)
            if folder.startswith(k):
                folder = v + folder[len(k):]
                break
        if absolute:
            return "/" + folder
        return folder

    @staticmethod
    def lowerize_dick_keys(thedict):
        return {k.lower(): v for k, v in thedict.items()}


class SonarrFilter(_FilterBase):
    def filter(self, request, res):
        # probably Sonarr
        low_request = self.lowerize_dick_keys(request)
        eventtype = low_request.get("eventtype")
        if eventtype == "Download":
            return self.process_download(request, res)
        return res.unhandled("Ignoring Sonarr event type: {!r}".format(eventtype))

    def process_download(self, request, res):
        log.debug("Processing Sonarr's 'on downloaded' event")
        res.update_status("processing")
        res.request_type = "sonarr"
        res.request_event = "on download"
        low_request = self.lowerize_dick_keys(request)
        low_series = self.lowerize_dick_keys(low_request.get("series", {}))
        root_dir = low_series.get("path")
        series_title = low_series.get("title")
        if not root_dir:
            return res.failed("Empty Series Path")
        root_dir = self.appy_path_mapping(root_dir)
        log.debug("Root folder: %s", root_dir)
        log.debug("Reconstructing full media path with basedir '%s'",
                  DopplerrConfig().get_cfg_value("general.basedir"))

        def concat_path(a, b):
            if not a.endswith('/'):
                a += '/'
            if b.startswith('/'):
                b = b[1:]
            a += b
            return a

        root_dir = concat_path(DopplerrConfig().get_cfg_value("general.basedir"), root_dir)
        basename = root_dir
        log.info("Searching episodes for serie '%s' in '%s'", series_title, root_dir)
        res.update_status("listing candidates")
        for episode in low_request.get("episodes", []):
            low_episode = self.lowerize_dick_keys(episode)
            basename = low_episode.get("scenename", "")
            episode_title = low_episode.get("title", "")
            season_number = low_episode.get("seasonnumber", "")
            episode_number = low_episode.get("episodenumber", "")
            quality = low_episode.get("quality", "")
            log.debug("Candidate: episode '%s' with base filename '%s'", episode_title, basename)
            res.candidates.append({
                "series_title": series_title,
                "episode_title": episode_title,
                "root_dir": root_dir,
                "scenename": basename,
                "season_number": season_number,
                "episode_number": episode_number,
                "quality": quality,
            })
        res.update_status("candidates listed")
        return res


class AutoDetectFilter(_FilterBase):
    def filter(self, request, res):
        low_request = self.lowerize_dick_keys(request)
        if "series" in low_request:
            return SonarrFilter().filter(request, res)
        return res.failed("Unable to find request type")



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

import treq

from twisted.internet.defer import inlineCallbacks

from dopplerr.cfg import DopplerrConfig

log = logging.getLogger(__name__)


class NotificationBase(object):

    NOTIFICATION_TYPES = ["fetched"]

    def __init__(self, registered_notification):
        self._registered_notification = registered_notification

    def can_emit_notification_type(self, requested_notif_type):
        return requested_notif_type in self._registered_notification


class NotificationPushOver(NotificationBase):

    __api_url = "https://api.pushover.net/1/messages.json"

    def __init__(self, token, user, registered_notifications):
        self.token = token
        self.user = user
        super(NotificationPushOver, self).__init__(registered_notifications)

    @inlineCallbacks
    def emit(self, notification_type, title, message):
        if not self.can_emit_notification_type(notification_type):
            log.debug("notification %s is ignored for pushover")
            return
        resp = yield treq.request(
            "POST",
            self.__api_url,
            json={
                "token": self.token,
                "user": self.user,
                "message": message,
                "title": title,
            })
        content = yield resp.text()
        log.debug("PushOver response: %r", content)


@inlineCallbacks
def emit_registered_notifications(notification_type, title, message):
    log.debug("Emiting notification: [%s] %s - %s", notification_type, title, message)
    if DopplerrConfig().get_cfg_value("notifications.pushover.enabled"):
        log.debug("Emiting pushover with user %s",
                  DopplerrConfig().get_cfg_value("notifications.pushover.user"))
        po = NotificationPushOver(
            DopplerrConfig().get_cfg_value("notifications.pushover.token"),
            DopplerrConfig().get_cfg_value("notifications.pushover.user"),
            DopplerrConfig().get_cfg_value("notifications.pushover.registered_notifications"),
        )
        yield po.emit(notification_type, title, message)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

log = logging.getLogger(__name__)


class Response(object):
    def __init__(self):
        self.res = {}
        self.update_status("unprocessed")

    def failed(self, message):
        log.error(message)
        self.res["status"] = "failed"
        self.res["message"] = message.lower()

    def unhandled(self, message):
        log.info("Filtered out event: %s", message)
        self.res["status"] = "unhandled"
        self.res["message"] = message.lower()

    @property
    def is_unhandled(self):
        return self.res.get("result", {}).get("status") == "unhandled"

    def update_status(self, status, message=None):
        self.res.setdefault("result", {})['status'] = status
        if message is not None:
            self.res['result']["message"] = message
        elif "message" in self.res:
            del self.res['result']["message"]

    @property
    def successful(self):
        return self.res.setdefault("result", {}).get("status") == "succeeded"

    def toDict(self):
        return self.res

    @property
    def request_type(self):
        return self.res.get("request", {}).get("type", None)

    @request_type.setter
    def request_type(self, tt):
        self.res.setdefault("request", {})["type"] = tt

    @property
    def request_event(self):
        return self.res.get("request", {}).get("event", None)

    @request_event.setter
    def request_event(self, ee):
        self.res.setdefault("request", {})["event"] = ee

    @property
    def exception(self):
        return self.res.get("result", {}).get("exception", None)

    @exception.setter
    def exception(self, ee):
        self.res.setdefault("result", {})["exception"] = ee

    @property
    def subtitles(self):
        return self.res.get("result", {}).get("subtitles", None)

    @subtitles.setter
    def subtitles(self, ee):
        self.res.setdefault("result", {})["subtitles"] = ee

    @property
    def candidates(self):
        return self.res.setdefault("candidates", [])

    @candidates.setter
    def candidates(self, ee):
        self.res.candidates = ee

    @property
    def sonarr_summary(self):
        candidate = self.candidates[0]
        subtitles = self.subtitles
        return [{
            "series_title": candidate.get("series_title"),
            "season_number": candidate.get("season_number"),
            "episode_number": candidate.get("episode_number"),
            "quality": candidate.get("quality"),
            "video_languages": candidate.get("video_languages", "???"),
            "subtitles_languages": ",".join(s["language"] for s in subtitles),
        }]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
from pathlib import Path

from txwebbackendbase.requests import dejsonify
from txwebbackendbase.requests import jsonify
from txwebbackendbase.singleton import singleton
from txwebbackendbase.threading import deferredAsThread

from klein import Klein
from twisted.internet.defer import inlineCallbacks
from twisted.web.static import File

from dopplerr import dopplerr_version
from dopplerr.cfg import DopplerrConfig
from dopplerr.db import DopplerrDb
from dopplerr.downloader import DopplerrDownloader
from dopplerr.notifications import emit_registered_notifications
from dopplerr.request_filter import SonarrFilter
from dopplerr.response import Response
from dopplerr.status import DopplerrStatus

log = logging.getLogger(__name__)


@singleton
class Routes(object):
    app = Klein()
    args = None
    __frontend_static = None
    __frontend_root = None

    def listen(self):
        DopplerrStatus().healthy = True
        self.app.run(host='0.0.0.0', port=int(DopplerrConfig().get_cfg_value("general.port")))

    @property
    def frontend_static(self):
        if self.__frontend_static is None:
            self.__frontend_static = str(
                Path(DopplerrConfig().get_cfg_value("general.frontenddir")) / "static")
        return self.__frontend_static

    @property
    def frontend_root(self):
        if self.__frontend_root is None:
            self.__frontend_root = str(Path(DopplerrConfig().get_cfg_value("general.frontenddir")))
        return self.__frontend_root

    # Double-index method kung-foo
    #   https://github.com/twisted/klein/issues/41
    # pylint: disable=function-redefined
    @app.route("/", branch=True)
    def index(self, _request):
        return File(self.frontend_root)

    @app.route("/", branch=True)
    def index(self, _request):
        return File(self.frontend_root)

    # pylint: enable=function-redefined

    @app.route("/static/", branch=True)
    def static(self, _request):
        return File(self.frontend_static)

    @app.route("/api/v1/recent/events/<num>")
    def recent_events_num(self, request, num=10):
        num = int(num)
        if num > 100:
            num = 100
        res = {"events": DopplerrDb().getRecentEvents(num)}
        return jsonify(request, res)

    @app.route("/api/v1/recent/events/")
    def recent_events_10(self, request):
        res = {"events": DopplerrDb().getRecentEvents(10)}
        return jsonify(request, res)

    @app.route("/api/v1/recent/fetched/series/<num>")
    def recent_fetched_series_num(self, request, num=10):
        num = int(num)
        if num > 100:
            num = 100
        res = {"events": DopplerrDb().getLastFetchedSeries(num)}
        return jsonify(request, res)

    @app.route("/api/v1/notify/sonarr", methods=['POST'])
    @inlineCallbacks
    def notify_sonarr(self, request):
        content = dejsonify(request)
        res = yield self._process_notify_sonarr(content)
        log.debug("Successful: %r", res.successful)
        if res.successful:
            for s in res.sonarr_summary:
                yield emit_registered_notifications(
                    "fetched", "Episode Subtitles Fetched",
                    ("{series_title} - {season_number}x{episode_number} - "
                     "{episode_title} [{quality}] - Lang: {video_languages} - "
                     "Subtitles: {subtitles_languages}".format(
                         series_title=s['series_title'],
                         season_number=s['season_number'],
                         episode_number=s['episode_number'],
                         episode_title=s['episode_number'],
                         quality=s['quality'],
                         video_languages=s['video_languages'],
                         subtitles_languages=s['subtitles_languages'],
                     )))
                DopplerrDb().insertFetchedSeriesSubtitles(
                    series_title=s['series_title'],
                    season_number=s['season_number'],
                    episode_number=s['episode_number'],
                    episode_title=s['episode_number'],
                    quality=s['quality'],
                    video_languages=s['video_languages'],
                    subtitles_languages=s['subtitles_languages'],
                )
        return jsonify(request, res.toDict())

    @deferredAsThread
    def _process_notify_sonarr(self, content):
        logging.debug("Notify sonarr request: %r", content)
        log.debug("Processing request: %r", content)
        res = Response()
        SonarrFilter().filter(content, res)
        if res.is_unhandled:
            # event has been filtered out
            return res
        candidates = res.candidates
        if not candidates:
            DopplerrDb().insertEvent("error", "event handled but no candidate found")
            log.debug("event handled but no candidate found")
            res.update_status("failed", "event handled but no candidate found")
            return res
        for candidate in candidates:
            log.info(
                "Searching episode '%s' from series '%s'. Filename: %s",
                candidate.get("episode_title"),
                candidate.get("series_title"),
                candidate.get("scenename"),
            )
            DopplerrDb().insertEvent("availability notification", "{} - {}x{} - {} [{}] available."
                                     .format(
                                         candidate.get("series_title"),
                                         candidate.get("season_number"),
                                         candidate.get("episode_number"),
                                         candidate.get("episode_title"),
                                         candidate.get("quality"),
                                     ))
            video_files_found = DopplerrDownloader().search_file(candidate['root_dir'],
                                                                 candidate['scenename'])
            log.debug("All found files: %r", video_files_found)
            if not video_files_found:
                res.update_status("failed", "candidates found but no video file found")
            else:
                DopplerrDownloader().download_missing_subtitles(res, video_files_found)
            subtitles = res.subtitles
            if not subtitles:
                DopplerrDb().insertEvent("subtitles", "not subtitles found for: {}"
                                         .format([Path(f).name for f in video_files_found]))
            else:
                DopplerrDb().insertEvent("subtitles", "subtitles fetched: {}".format(
                    ", ".join([
                        "{} (lang: {}, source: {})".format(
                            s.get("filename"),
                            s.get("language"),
                            s.get("provider"),
                        ) for s in subtitles
                    ])))

        return res

    @app.route("/api/v1/notify", methods=['GET'])
    def notify_not_allowed(self, request):
        request.setResponseCode(405)
        return "Method GET not allowed. Use POST with a JSON body with the right format"

    @app.route("/api/v1/health")
    def health(self, request):
        res_health = {
            "healthy": DopplerrStatus().healthy,
            "languages": DopplerrConfig().get_cfg_value("subliminal.languages"),
            "mapping": DopplerrConfig().get_cfg_value("general.mapping"),
            "version": dopplerr_version,
        }
        return jsonify(request, res_health)

    @app.route("/api/v1/version")
    def version(self, _request):
        return dopplerr_version

    @app.route("/api/v1/fullscan")
    @inlineCallbacks
    def fullscan(self, request):
        content = dejsonify(request)
        logging.debug("Fullscan request: %r", content)
        res = yield DopplerrDownloader().process_fullscan(content)
        return jsonify(request, res)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

from txwebbackendbase.singleton import singleton

from dopplerr.cfg import DopplerrConfig

log = logging.getLogger(__name__)


@singleton
class DopplerrStatus(object):
    """
    Contains current status of the application and derived values from DopplerrConfig
    """

    def __init__(self):
        self.healthy = False
        self.sqlite_db_path = None
        self.subliminal_provider_configs = None

    def refresh_from_cfg(self):
        """
        Refresh derived values from cfg
        """
        cfg = DopplerrConfig()
        if not cfg.get_cfg_value("general.port"):
            log.fatal("No port defined !")
            raise Exception("No port defined")
        if not cfg.get_cfg_value("general.frontenddir"):
            log.fatal("No frontend dir defined")
            raise Exception("No frontend dir defined")
        self.subliminal_provider_configs = self._build_subliminal_provider_cfgs()

        languages = cfg.get_cfg_value("subliminal.languages")
        if not languages:
            raise Exception("No languages defined")
        if any(not x for x in languages):
            raise Exception("Bad languages: {!r}".format(languages))

    def _build_subliminal_provider_cfgs(self):
        cfg = DopplerrConfig()
        provider_configs = {}
        provider_names = [
            "addic7ed",
            "legendastv",
            "opensubtitles",
            "subscenter",
        ]
        for provider_name in provider_names:
            if cfg.get_cfg_value("subliminal.{}.enabled".format(provider_name)):
                provider_configs[provider_name] = {
                    'username': cfg.get_cfg_value("subliminal.{}.user".format(provider_name)),
                    'password': cfg.get_cfg_value("subliminal.{}.password".format(provider_name)),
                }
                log.debug("Using %s username: %s", provider_name,
                          provider_configs[provider_name]['username'])
        return provider_configs



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import pbr.version

__version__ = pbr.version.VersionInfo('dopplerr').release_string()
version = __version__
dopplerr_version = __version__

__all__ = [
    '__version__',
    'dopplerr_version',
    'version',
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import logging.config
import os
import sys
from pathlib import Path

from babelfish import Language
from txwebbackendbase.logging import setupLogger

from dopplerr import dopplerr_version
from dopplerr.cfg import DopplerrConfig
from dopplerr.db import DopplerrDb
from dopplerr.downloader import DopplerrDownloader
from dopplerr.routes import Routes
from dopplerr.status import DopplerrStatus

log = logging.getLogger(__name__)


def print_list(aList):
    return ",".join(aList)


def list_of_languages(langList):
    langs = [s.lower() for s in langList.split(',')]
    failed = False
    for l in langs:
        try:
            Language(l)
        except ValueError:
            failed = True
            logging.critical("Invalid language: %r", l)
    if failed:
        sys.exit(2)
    return langs


def main():
    if "-v" in sys.argv or "--general-verbose" in sys.argv:
        default_level = logging.DEBUG
    else:
        default_level = logging.INFO
    setupLogger(level=default_level, no_color=True)
    log.debug("Initializing Dopplerr version %s...", dopplerr_version)

    DopplerrConfig().find_configuration_values()

    setupLogger(
        level=(logging.DEBUG
               if DopplerrConfig().get_cfg_value("general.verbose") else logging.WARNING),
        no_color=DopplerrConfig().get_cfg_value("general.no_color"),
        logfile=DopplerrConfig().get_cfg_value("general.logfile"))
    log.info("Reset logging format to %s", "verbose"
             if DopplerrConfig().get_cfg_value("general.verbose") else "not verbose")

    log.debug("Applying configuration")
    DopplerrStatus().refresh_from_cfg()

    log.info("Initializing Subtitle DopplerrDownloader Service")

    DopplerrDownloader().initialize_subliminal()
    DopplerrStatus().sqlite_db_path = (
        Path(DopplerrConfig().get_cfg_value("general.configdir")) / "sqlite.db")
    log.debug("SQLite DB: %s", DopplerrStatus().sqlite_db_path.as_posix())
    DopplerrDb().init(DopplerrStatus().sqlite_db_path)
    DopplerrDb().createTables()

    # change current work dir for subliminal work files
    os.chdir(DopplerrConfig().get_cfg_value("general.configdir"))

    DopplerrDb().insertEvent("start", "dopplerr started")

    # main event loop (Twisted reactor behind)
    Routes().listen()

    DopplerrDb().insertEvent("stop", "dopplerr stopped")



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
# import sqlite3
import threading
from pathlib import Path

# from peewee import ForeignKeyField
from peewee import BooleanField
from peewee import CharField
from peewee import DateTimeField
from peewee import IntegerField
from peewee import Model
from peewee import SqliteDatabase
from peewee import TextField
from peewee import Using
from txwebbackendbase.singleton import singleton


class Events(Model):
    timestamp = DateTimeField(default=datetime.datetime.now)
    type = CharField()
    message = TextField()


class MissingSubtitles(Model):
    path = CharField()
    found = BooleanField()


class FetchedSeriesSubtitles(Model):
    timestamp = DateTimeField(default=datetime.datetime.now)
    series_title = CharField()
    season_number = IntegerField()
    episode_number = IntegerField()
    episode_title = CharField()
    quality = CharField()
    video_languages = CharField()
    subtitles_languages = CharField()


@singleton
class DopplerrDb(object):
    def __init__(self):
        self.__sqlite_db_path = None
        self.__conn = None
        self.__database = None
        self.__db_lock = threading.Lock()

    @property
    def database(self):
        if not self.__database:
            self.__database = SqliteDatabase(self.__sqlite_db_path)
        return self.__database

    @property
    def conn(self):
        return self.database.get_conn()

    def init(self, sqlite_db_path: Path):
        self.__sqlite_db_path = sqlite_db_path.as_posix()

    def createTables(self):
        self.database.create_table(MissingSubtitles, safe=True)
        self.database.create_table(Events, safe=True)
        self.database.create_table(FetchedSeriesSubtitles, safe=True)

    def insertEvent(self, thetype: str, message: str):
        with Using(self.database, [Events], with_transaction=False):
            Events.insert(type=thetype, message=message).execute()

    def getRecentEvents(self, limit: int):
        with Using(self.database, [Events], with_transaction=False):
            events = (Events.select().limit(limit).order_by(Events.timestamp.desc()).execute())
            return [{
                "timestamp": e.timestamp.strftime('%Y-%m-%dT%H:%M:%S'),
                "type": e.type,
                "message": e.message
            } for e in events]

    def insertFetchedSeriesSubtitles(self, series_title, season_number, episode_number,
                                     episode_title, quality, video_languages, subtitles_languages):
        with Using(self.database, [FetchedSeriesSubtitles], with_transaction=True):
            FetchedSeriesSubtitles.insert(
                series_title=series_title,
                season_number=season_number,
                episode_number=episode_number,
                episode_title=episode_title,
                quality=quality,
                video_languages=video_languages,
                subtitles_languages=subtitles_languages,
            ).execute()

    def getLastFetchedSeries(self, limit: int):
        with Using(self.database, [FetchedSeriesSubtitles], with_transaction=False):
            events = (FetchedSeriesSubtitles.select().limit(limit).order_by(
                FetchedSeriesSubtitles.timestamp.desc()).execute())
            return [{
                "timestamp": e.timestamp.strftime('%Y-%m-%dT%H:%M:%S'),
                "series_title": e.series_title,
                "season_number": e.season_number,
                "episode_number": e.episode_number,
                "episode_title": e.episode_title,
                "quality": e.quality,
                "video_languages": e.video_languages,
                "subtitles_languages": e.subtitles_languages,
            } for e in events]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import os
import threading

# from subliminal import scan_videos
from babelfish import Language
from subliminal import Video
from subliminal import download_best_subtitles
from subliminal import region
from subliminal import save_subtitles
from subliminal.subtitle import get_subtitle_path
from txwebbackendbase.singleton import singleton
from txwebbackendbase.utils import recursive_iglob

from dopplerr.cfg import DopplerrConfig
from dopplerr.status import DopplerrStatus

log = logging.getLogger(__name__)


@singleton
class DopplerrDownloader(object):
    def __init__(self):
        # Avoid having 2 download at the same time, at least of the integrity of the cache file
        self.subliminal_download_lock = threading.Lock()

    @staticmethod
    def initialize_subliminal():
        log.info("Initializing Subliminal cache...")
        region.configure('dogpile.cache.dbm', arguments={'filename': 'cachefile.dbm'})

    def search_file(self, root_dir, base_name):
        # This won't work with python < 3.5
        found = []
        protected_path = os.path.join(root_dir, "**", "*" + base_name + "*")
        protected_path = protected_path.replace("[", "[[]").replace("]", "[]]")
        log.debug("Searching %r", protected_path)
        for filename in recursive_iglob(protected_path):
            log.debug("Found: %s", filename)
            found.append(filename)
        return found

    def process_fullscan(self, _request):
        log.debug("Processing full scan of missing subtitle files...")
        res = {
            'status': 'unprocessed',
            'message': 'not implemented yet!',
        }
        # TODO: inspiration
        #   https://gist.github.com/alexsavio/9299716
        return res

    def download_missing_subtitles(self, res, files):
        log.info("Searching and downloading missing subtitles for: %r", files)
        res.update_status("downloading", "downloading missing subtitles")
        videos = []
        for fil in files:
            _, ext = os.path.splitext(fil)
            if ext in [".jpeg", ".jpg", ".nfo", ".srt", ".sub", ".nbz"]:
                log.debug("Ignoring %s because of extension: %s", fil, ext)
                continue
            videos.append(Video.fromname(fil))
        log.info("Video files: %r", videos)
        if not videos:
            log.debug("No subtitle to download")
            res.update_status("failed", "no video file found")
            return res
        res.update_status("fetching", "finding best subtitles")
        self.subliminal_download_lock.acquire()
        log.info("fetching subtitles...")
        try:
            provider_configs = DopplerrStatus().subliminal_provider_configs
            subtitles = download_best_subtitles(
                videos,
                {Language(l)
                 for l in DopplerrConfig().get_cfg_value("subliminal.languages")},
                provider_configs=provider_configs)
        except Exception as e:
            log.exception("subliminal raised an exception")
            res.update_status("failed", "subliminal exception")
            res.exception = repr(e)
            return res
        self.subliminal_download_lock.release()
        subtitles_info = []
        for vid in videos:
            log.info("Found subtitles for %s:", vid)
            for sub in subtitles[vid]:
                log.info("  %s from %s", sub.language, sub.provider_name)
                subtitles_info.append({
                    "language": str(sub.language),
                    "provider": sub.provider_name,
                    "filename": get_subtitle_path(vid.name, language=sub.language)
                })
            save_subtitles(vid, subtitles[vid])
        res.subtitles = subtitles_info
        if subtitles_info:
            res.update_status("succeeded", "download successful")
        else:
            res.update_status("failed", "no subtitle found")
        return res



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import os
from pathlib import PosixPath

import pkg_resources
from txwebbackendbase.singleton import singleton

from dopplerr.environment_config import BoolCfg
from dopplerr.environment_config import ConfigFileCfg
from dopplerr.environment_config import DirNameCfg
from dopplerr.environment_config import EnvironmentConfig
from dopplerr.environment_config import IntCfg
from dopplerr.environment_config import JsonFileConfigEntryPoint
from dopplerr.environment_config import ListOfStringCfg
from dopplerr.environment_config import MultiChoiceCfg
from dopplerr.environment_config import PasswordCfg
from dopplerr.environment_config import StringCfg
from dopplerr.environment_config import UserCfg

log = logging.getLogger(__name__)
DEFAULT_PORT = 8086


def _find_frontend_data():
    installed_data_frontend = pkg_resources.resource_filename(__name__, 'frontend')
    if PosixPath(installed_data_frontend).exists():
        log.debug("Found local frontend path: %s", installed_data_frontend)
        return installed_data_frontend
    setup_py = pkg_resources.resource_filename(__name__, "main.py")
    dev_env_frontend_dist = PosixPath(setup_py).parent.parent / "frontend" / "dist"
    if dev_env_frontend_dist.exists():
        log.debug("Found dev local frontend path: %s", dev_env_frontend_dist)
        return str(dev_env_frontend_dist)
    return None


class DopplerrJsonConfigFile(JsonFileConfigEntryPoint):
    _CONFIG_ENTRYPOINT_ENVIRON_VAR_NAME = "DOPPLERR_COMMON_CONFIG_FILE"
    _CONFIG_ENTRYPOINT_LONG_PARAM_NAME = "--configfile"
    _CONFIG_ENTRYPOINT_SHORT_PARAM_NAME = "-g"
    _CONFIG_ENTRYPOINY_DEFAULT_FILENAME = "config.json"


@singleton
class DopplerrConfig(EnvironmentConfig):

    _ENVIRON_VAR_PREFIX = "DOPPLERR_"
    _CONFIGURATION_ENTRY_POINT = DopplerrJsonConfigFile()

    _CFG = {
        "configfile": ConfigFileCfg(l="--configfile", h="Config directory"),
        "general": {
            "basedir":
                DirNameCfg(s="-b", d=os.getcwd(), h='Base directory'),
            "configdir":
                DirNameCfg(s="-c", d=os.getcwd(), h="Config directory"),
            "appdir":
                DirNameCfg(s="-a", d=os.getcwd(), h="App directory"),
            "frontenddir":
                DirNameCfg(s="-f", d=_find_frontend_data(), r=True, h="Frontend directory"),
            "verbose":
                BoolCfg(s='-v', h='Enable verbose output'),
            "logfile":
                StringCfg(s="-l", h='Output log to file'),
            "mapping":
                ListOfStringCfg(
                    s='-m',
                    h=("Map root folder of tv/anime/movie to another name.\n"
                       "Ex: series are mounted on a docker image as /tv but \n"
                       "on the other system it is under /video/Series. In this \n"
                       "case use '--basedir /video --mapping tv=Series,movies=Movies'")),
            "port":
                IntCfg(s='-p', d=DEFAULT_PORT, h='The port to listen on'),
            "no_color":
                BoolCfg(h="Disable color in logs"),
        },
        "subliminal": {
            "languages": ListOfStringCfg(),
            "addic7ed": {
                "enabled": BoolCfg(h="Enable addic7ed"),
                "user": UserCfg(h="addic7ed username"),
                "password": PasswordCfg(h="addic7ed password"),
            },
            "legendastv": {
                "enabled": BoolCfg(h="Enable legendastv"),
                "user": UserCfg(h="legendastv username"),
                "password": PasswordCfg(h="legendastv password"),
            },
            "opensubtitles": {
                "enabled": BoolCfg(h="Enable opensubtitles"),
                "user": UserCfg(h="opensubtitles username"),
                "password": PasswordCfg(h="opensubtitles password"),
            },
            "subscenter": {
                "enabled": BoolCfg(h="Enable subscenter"),
                "user": UserCfg(h="subscenter username"),
                "password": PasswordCfg(h="subscenter password"),
            },
        },
        "notifications": {
            "pushover": {
                "enabled":
                    BoolCfg(h="Enable pushover"),
                "user":
                    UserCfg(h="pushover username"),
                "token":
                    PasswordCfg(h="pushover password"),
                "registered_notifications":
                    MultiChoiceCfg(h="Registerd Notifications", choices=["fetched"], d=["fetched"]),
            }
        }
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import setuptools

# In python < 2.7.4, a lazy loading of package `pbr` will break
# setuptools if some other modules registered functions in `atexit`.
# solution from: http://bugs.python.org/issue15881#msg170215
try:
    import multiprocessing  # noqa
except ImportError:
    pass

setuptools.setup(setup_requires=['pbr'],
                 pbr=True)
